package org.netkernel.ext.system.accessor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.RandomAccessFile;

import org.netkernel.layer0.logging.XMLFormatter;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

public class ViewLogAccessor extends StandardAccessorImpl
{
	private static final int LINE_MOD=10;
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		INKFRequestReadOnly thisReq=aContext.getThisRequest();
		String logId=thisReq.getArgumentValue("log");
		int offset=Integer.parseInt(thisReq.getArgumentValue("offset"));
		int count=Integer.parseInt(thisReq.getArgumentValue("count"));
		
		IHDSNode logs=aContext.source("netkernel:/logLevels",IHDSNode.class);
		IHDSNode log=logs.getNodes("/logs/log").filter(HDSPredicateFactory.namedChildEquals("id", logId)).getFirstNode();
		if (log==null)
		{	throw new NKFException("log not found");
		}
		String filename=log.getFirstValue("file").toString();
		if (filename==null)
		{	throw new NKFException("log format not suitable for viewing");
		}
		
		File f = new File(filename);
		RandomAccessFile raf = new RandomAccessFile(f,"r");
		StringBuffer sb;
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream(4096);
		OutputStreamWriter osw = new OutputStreamWriter(baos);
		osw.write("<log>");
		osw.flush();
		
		boolean atStart=writeLastLines((count+offset)*LINE_MOD,count*LINE_MOD,baos,raf);
		raf.close();
		if (atStart)
		{	osw.write("<atStart/>");
		}
		osw.write("</log>");
		osw.flush();
		
		INKFResponse resp=aContext.createResponseFrom(new ByteArrayRepresentation(baos));
		resp.setMimeType("text/xml");
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	
	private boolean writeLastLines(int aOffset, int aCount, OutputStream aStream, RandomAccessFile aFile) throws Exception
	{
		final int  BUFSIZE = 1024;
		long pos, endPos, lastPos, lastPos2;
		int lineCounter = 0;
		byte[] buffer = new byte[BUFSIZE];
		RandomAccessFile in = aFile;
		boolean atStart=false;
		
		endPos = in.length(); // determine file length
		lastPos = endPos; // initialize lastPos to end of file
		lastPos2 = endPos;

		// guarantee a non-empty file before checking if the very
		// last character is a newline, count one if not
		if (endPos > 0)
		{
			byte[] oneByte = new byte[1];
			in.seek(endPos-1);
			in.read(oneByte);
			if ((char)oneByte[0] != '\n') lineCounter++;
		}

		do
		{
			// seek cursor back BUFSIZE bytes or to beginning of
			// file if there aren't BUFSIZE bytes remaining
			pos = ((lastPos-BUFSIZE) > 0) ? lastPos - BUFSIZE : 0;
			in.seek(pos);

			// if there aren't BUFSIZE bytes left, adjust buffer to
			// read the remaining bytes
			if ((lastPos-pos) < BUFSIZE)
				buffer = new byte[(int)(lastPos-pos)];

			// read bytes in to buffer
			in.readFully(buffer);

			// read backwards through the buffer counting newlines
			for (int j = buffer.length-1; j >= 0; j--)
			{
				if ((char)buffer[j] == '\n')
				{
					if ((lineCounter%LINE_MOD)==0) lastPos2 = pos+j+1;
					lineCounter++;
					if (lineCounter > aOffset)
					{
						pos += (j+1);
						break;
					}
				}
			}

			// update lastPos before looping
			lastPos = pos;
		} while((lineCounter <= aOffset) && (pos != 0));
		if (lineCounter<aOffset)
		{	lineCounter=(lineCounter/LINE_MOD)*LINE_MOD;
			aCount-=(aOffset-lineCounter);
			if (aCount<0)
			{	aCount=0;
			}
			atStart=true;
			lastPos2=lastPos;
		}
		
		int lines=0;
		int n=0;
		pos=lastPos2;
		while(lines<aCount)
		{	in.seek(pos);
			n=in.read(buffer);
			if (n>0)
			{	for (int i=0; i<n; i++)
				{	if (buffer[i]==XMLFormatter.NEWLINE_SUBSTITUTE)
					{	buffer[i]='\n';
					}
					else if (buffer[i]=='\t')
					{	buffer[i]=' ';
					}
					else if (buffer[i]=='\n')
					{	lines++;
						if (lines==aCount)
						{	n=i+1;
						}
					}
				}
				aStream.write(buffer,0,n);
				pos+=n;
			}
			else
			{	break;
			}
		}
		return atStart;
			
	}	
}
